import json
import math
import shutil
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from contextlib import suppress
from functools import cached_property
from pathlib import Path, PurePosixPath
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    overload,
)

import numpy as np
import polars as pl
from filelock import FileLock
from loguru import logger
from rich.progress import track
from semver.version import Version
from typing_extensions import Self, override

from luxonis_ml.data.utils import (
    BucketStorage,
    BucketType,
    ParquetFileManager,
    UpdateMode,
    get_class_distributions,
    get_duplicates_info,
    get_heatmaps,
    get_missing_annotations,
    infer_task,
    warn_on_duplicates,
)
from luxonis_ml.data.utils.constants import LDF_VERSION
from luxonis_ml.enums.enums import DatasetType
from luxonis_ml.typing import PathType
from luxonis_ml.utils import (
    LuxonisFileSystem,
    deprecated,
    environ,
    make_progress_bar,
)

from .annotation import (
    Category,
    DatasetRecord,
    Detection,
)
from .base_dataset import BaseDataset, DatasetIterator
from .metadata import Metadata
from .migration import migrate_dataframe, migrate_metadata
from .source import LuxonisSource
from .utils import find_filepath_uuid, get_dir, get_file


class LuxonisDataset(BaseDataset):
    def __init__(
        self,
        dataset_name: str,
        team_id: Optional[str] = None,
        bucket_type: Union[
            BucketType, Literal["internal", "external"]
        ] = BucketType.INTERNAL,
        bucket_storage: Union[
            BucketStorage, Literal["local", "gcs", "s3", "azure"]
        ] = BucketStorage.LOCAL,
        *,
        delete_local: bool = False,
        delete_remote: bool = False,
    ) -> None:
        """Luxonis Dataset Format (LDF) is used to define datasets in
        the Luxonis MLOps ecosystem.

        @type dataset_name: str
        @param dataset_name: Name of the dataset
        @type team_id: Optional[str]
        @param team_id: Optional unique team identifier for the cloud
        @type bucket_type: BucketType
        @param bucket_type: Whether to use external cloud buckets
        @type bucket_storage: BucketStorage
        @param bucket_storage: Underlying bucket storage. Can be one of
            C{local}, C{S3}, or C{GCS}.
        @type delete_local: bool
        @param delete_local: Whether to delete a dataset with the same
            name if it exists
        @type delete_remote: bool
        @param delete_remote: Whether to delete the dataset from the
            cloud as well
        """

        self.dataset_name = dataset_name
        self.base_path = environ.LUXONISML_BASE_PATH
        self.base_path.mkdir(exist_ok=True)

        self._credentials = self._init_credentials()
        self._is_synced = False

        # What is this for?
        self.bucket_type = BucketType(bucket_type)

        self.bucket_storage = BucketStorage(bucket_storage)

        if self.bucket_storage == BucketStorage.AZURE_BLOB:
            raise NotImplementedError("Azure Blob Storage not yet supported")

        self.bucket = self._get_credential("LUXONISML_BUCKET")

        if self.is_remote and self.bucket is None:
            raise ValueError(
                "The `LUXONISML_BUCKET` environment variable "
                "must be set for remote datasets"
            )

        self.team_id = team_id or self._get_credential("LUXONISML_TEAM_ID")

        self._init_paths()

        self.fs = LuxonisFileSystem(self.path)

        if delete_local or delete_remote:
            if self.exists(
                self.dataset_name,
                self.team_id,
                self.bucket_storage,
                self.bucket,
            ):
                self.delete_dataset(
                    delete_remote=delete_remote, delete_local=delete_local
                )

            self._init_paths()

        # For DDP GCS training - multiple processes
        with FileLock(self.base_path / ".metadata.lock"):
            self._metadata = self._get_metadata()

        if self.version != LDF_VERSION:
            logger.warning(
                f"LDF versions do not match. The current `luxonis-ml` "
                f"installation supports LDF v{LDF_VERSION}, but the "
                f"`{self.identifier}` dataset is in v{self._metadata.ldf_version}. "
                "Internal migration will be performed. Note that some parts "
                "and new features might not work correctly unless you "
                "manually re-create the dataset using the latest version "
                "of `luxonis-ml`."
            )
        self.progress = make_progress_bar()

    @property
    def metadata(self) -> Metadata:
        """Returns a copy of the dataset metadata.

        The metadata is a pydantic model with the following fields:
            - source: L{LuxonisSource}
            - ldf_version: str
            - classes: Dict[task_name, Dict[class_name, class_id]]
            - tasks: Dict[task_name, List[task_type]]
            - skeletons: Dict[task_name, Skeletons]
              - Skeletons is a dictionary with keys 'labels' and 'edges'
                - labels: List[str]
                - edges: List[Tuple[int, int]]
            - categorical_encodings: Dict[task_name, Dict[metadata_name, Dict[metadata_value, int]]]
              - Encodings for string metadata values
              - Example::

                    {
                        "vehicle": {
                            "color": {"red": 0, "green": 1, "blue": 2},
                            "brand": {"audi": 0, "bmw": 1, "mercedes": 2},
                        }
                    }

        @type: L{Metadata}
        """
        return self._metadata.model_copy(deep=True)

    @cached_property
    def version(self) -> Version:
        return Version.parse(
            self._metadata.ldf_version, optional_minor_and_patch=True
        )

    @property
    def source(self) -> LuxonisSource:
        if self._metadata.source is None:
            raise ValueError("Source not found in metadata")
        return self._metadata.source

    @property
    @override
    def identifier(self) -> str:
        return self.dataset_name

    def __len__(self) -> int:
        """Returns the number of instances in the dataset."""
        if self.is_remote:
            return len(list(self.fs.walk_dir("media")))

        df = self._load_df_offline()
        return len(df.select("uuid").unique()) if df is not None else 0

    def _get_credential(self, key: str) -> str:
        """Gets secret credentials from credentials file or ENV
        variables."""

        if key in self._credentials:
            return self._credentials[key]
        if not hasattr(environ, key):
            raise RuntimeError(f"Must set {key} in ENV variables")
        return getattr(environ, key)

    def _init_paths(self) -> None:
        """Configures local path or bucket directory."""

        self.local_path = (
            self.base_path
            / "data"
            / self.team_id
            / "datasets"
            / self.dataset_name
        )
        self.media_path = self.local_path / "media"
        self.annotations_path = self.local_path / "annotations"
        self.metadata_path = self.local_path / "metadata"
        self.arrays_path = self.local_path / "arrays"

        for path in [
            self.media_path,
            self.annotations_path,
            self.metadata_path,
        ]:
            path.mkdir(exist_ok=True, parents=True)

        if not self.is_remote:
            self.path = str(self.local_path)
        else:
            self.path = self._construct_url(
                self.bucket_storage,
                self.bucket,
                self.team_id,
                self.dataset_name,
            )

    def _save_df_offline(self, pl_df: pl.DataFrame) -> None:
        """Saves the given Polars DataFrame into multiple Parquet files
        using ParquetFileManager. Ensures the same structure as the
        original dataset.

        @type pl_df: pl.DataFrame
        @param pl_df: The Polars DataFrame to save.
        """
        annotations_path = Path(self.annotations_path)

        for old_file in annotations_path.glob("*.parquet"):
            old_file.unlink()

        rows = pl_df.to_dicts()

        with ParquetFileManager(annotations_path) as pfm:
            for row in rows:
                uuid_val = row.get("uuid")
                if uuid_val is None:
                    raise ValueError("Missing 'uuid' in row!")

                data_dict = dict(row)
                data_dict.pop("uuid", None)

                pfm.write(uuid_val, data_dict)  # type: ignore

        logger.info(
            f"Saved merged DataFrame to Parquet files in '{annotations_path}'."
        )

    def _merge_metadata_with(self, other: "LuxonisDataset") -> None:
        """Merges relevant metadata from `other` into `self`."""
        self._metadata = self._metadata.merge_with(other._metadata)
        self._write_metadata()

    def clone(
        self, new_dataset_name: str, push_to_cloud: bool = True
    ) -> "LuxonisDataset":
        """Create a new LuxonisDataset that is a local copy of the
        current dataset. Cloned dataset will overwrite the existing
        dataset with the same name.

        @type new_dataset_name: str
        @param new_dataset_name: Name of the newly created dataset.
        @type push_to_cloud: bool
        @param push_to_cloud: Whether to push the new dataset to the
            cloud. Only if the current dataset is remote.
        """

        new_dataset = LuxonisDataset(
            dataset_name=new_dataset_name,
            team_id=self.team_id,
            bucket_type=self.bucket_type,
            bucket_storage=self.bucket_storage,
            delete_local=True,
            delete_remote=True,
        )

        if self.is_remote:
            self.pull_from_cloud(update_mode=UpdateMode.MISSING)

        new_dataset_path = Path(new_dataset.local_path)
        new_dataset_path.mkdir(parents=True, exist_ok=True)
        shutil.copytree(
            self.local_path, new_dataset.local_path, dirs_exist_ok=True
        )

        new_dataset._init_paths()
        new_dataset._metadata = self._get_metadata()

        new_dataset._metadata.parent_dataset = self.dataset_name

        if push_to_cloud:
            if self.is_remote:
                new_dataset.push_to_cloud(
                    update_mode=UpdateMode.MISSING,
                    bucket_storage=self.bucket_storage,
                )
            else:
                logger.warning(
                    f"Cannot push to cloud. The cloned dataset '{new_dataset.dataset_name}' is local. "
                )

        new_dataset._write_metadata()

        return new_dataset

    def merge_with(
        self,
        other: "LuxonisDataset",
        inplace: bool = True,
        new_dataset_name: Optional[str] = None,
    ) -> "LuxonisDataset":
        """Merge all data from `other` LuxonisDataset into the current
        dataset (in-place or in a new dataset).

        @type other: LuxonisDataset
        @param other: The dataset to merge into the current dataset.
        @type inplace: bool
        @param inplace: Whether to merge into the current dataset (True)
            or create a new dataset (False).
        @type new_dataset_name: str
        @param new_dataset_name: The name of the new dataset to create
            if inplace is False.
        """
        if inplace:
            target_dataset = self
        elif new_dataset_name:
            if self.bucket_storage != other.bucket_storage:
                raise ValueError(
                    "Cannot merge datasets with different bucket storage types."
                )
            target_dataset = self.clone(new_dataset_name, push_to_cloud=False)
        else:
            raise ValueError(
                "You must specify a name for the new dataset "
                "when inplace is False"
            )

        if self.is_remote:
            other.pull_from_cloud(UpdateMode.MISSING)
            self.pull_from_cloud(UpdateMode.MISSING)

        df_self = self._load_df_offline(raise_when_empty=True)
        df_other = other._load_df_offline(raise_when_empty=True)
        duplicate_uuids = set(df_self["uuid"]).intersection(df_other["uuid"])
        if duplicate_uuids:
            logger.warning(
                f"Found {len(duplicate_uuids)} duplicate UUIDs in the datasets. "
                "Merging will remove these duplicates from the incoming dataset."
            )
            df_other = df_other.filter(
                ~df_other["uuid"].is_in(duplicate_uuids)
            )

        df_merged = pl.concat([df_self, df_other])
        target_dataset._save_df_offline(df_merged)

        splits_self = self._load_splits(self.metadata_path)
        splits_other = self._load_splits(
            other.metadata_path
        )  # dict of split names to list of uuids
        splits_other = {
            split_name: [uuid for uuid in uuids if uuid not in duplicate_uuids]
            for split_name, uuids in splits_other.items()
        }
        self._merge_splits(splits_self, splits_other)
        target_dataset._save_splits(splits_self)

        if self.is_remote:
            shutil.copytree(
                other.media_path, target_dataset.media_path, dirs_exist_ok=True
            )
            target_dataset.push_to_cloud(
                bucket_storage=target_dataset.bucket_storage,
                update_mode=UpdateMode.MISSING,
            )

        target_dataset._merge_metadata_with(other)

        return target_dataset

    def _load_splits(self, path: Path) -> Dict[str, List[str]]:
        splits_path = path / "splits.json"
        with open(splits_path) as f:
            return json.load(f)

    def _merge_splits(
        self,
        splits_self: Dict[str, List[str]],
        splits_other: Dict[str, List[str]],
    ) -> None:
        for split_name, uuids_other in splits_other.items():
            if split_name not in splits_self:
                splits_self[split_name] = []
            combined_uuids = set(splits_self[split_name]).union(uuids_other)
            splits_self[split_name] = list(combined_uuids)

    def _save_splits(self, splits: Dict[str, List[str]]) -> None:
        splits_path_self = self.metadata_path / "splits.json"
        with open(splits_path_self, "w") as f:
            json.dump(splits, f, indent=4)

    @overload
    def _load_df_offline(
        self,
        lazy: Literal[False] = ...,
        raise_when_empty: Literal[False] = ...,
        attempt_migration: bool = ...,
    ) -> Optional[pl.DataFrame]: ...

    @overload
    def _load_df_offline(
        self,
        lazy: Literal[False] = ...,
        raise_when_empty: Literal[True] = ...,
        attempt_migration: bool = ...,
    ) -> pl.DataFrame: ...

    @overload
    def _load_df_offline(
        self,
        lazy: Literal[True] = ...,
        raise_when_empty: Literal[False] = ...,
        attempt_migration: bool = ...,
    ) -> Optional[pl.LazyFrame]: ...

    @overload
    def _load_df_offline(
        self,
        lazy: Literal[True] = ...,
        raise_when_empty: Literal[True] = ...,
        attempt_migration: bool = ...,
    ) -> pl.LazyFrame: ...

    def _load_df_offline(
        self,
        lazy: bool = False,
        raise_when_empty: bool = False,
        attempt_migration: bool = True,
    ) -> Optional[Union[pl.DataFrame, pl.LazyFrame]]:
        """Loads the dataset DataFrame **always** from the local
        storage."""
        path = (
            self.base_path
            / "data"
            / self.team_id
            / "datasets"
            / self.dataset_name
            / "annotations"
        )

        if not path.exists():
            if raise_when_empty:
                raise FileNotFoundError(
                    f"Dataset '{self.dataset_name}' is empty."
                )
            return None

        if lazy:
            dfs = [pl.scan_parquet(file) for file in path.glob("*.parquet")]
            df = pl.concat(dfs) if dfs else None
        else:
            dfs = [pl.read_parquet(file) for file in path.glob("*.parquet")]
            df = pl.concat(dfs) if dfs else None

        if df is None and raise_when_empty:
            raise FileNotFoundError(f"Dataset '{self.dataset_name}' is empty.")

        if not attempt_migration or self.version == LDF_VERSION or df is None:
            return df
        return migrate_dataframe(df)  # pragma: no cover

    @overload
    def _get_index(
        self,
        lazy: Literal[False] = ...,
        raise_when_empty: Literal[False] = ...,
    ) -> Optional[pl.DataFrame]: ...

    @overload
    def _get_index(
        self,
        lazy: Literal[False] = ...,
        raise_when_empty: Literal[True] = ...,
    ) -> pl.DataFrame: ...

    @overload
    def _get_index(
        self,
        lazy: Literal[True] = ...,
        raise_when_empty: Literal[False] = ...,
    ) -> Optional[pl.LazyFrame]: ...

    def _get_index(
        self,
        lazy: bool = False,
        raise_when_empty: bool = False,
    ) -> Optional[Union[pl.DataFrame, pl.LazyFrame]]:
        """Loads unique file entries from annotation data."""
        df = self._load_df_offline(
            lazy=lazy, raise_when_empty=raise_when_empty
        )
        if df is None:
            return None

        processed = df.select(
            pl.col("uuid"),
            pl.col("file").str.extract(r"([^\/\\]+)$").alias("file"),
            pl.col("file")
            .apply(lambda x: str(Path(x).resolve()), return_dtype=pl.Utf8)
            .alias("original_filepath"),
        )

        processed = processed.unique(
            subset=["uuid", "original_filepath"], maintain_order=False
        )

        if not lazy and isinstance(processed, pl.LazyFrame):
            processed = processed.collect()

        return processed

    def _write_metadata(self) -> None:
        path = self.metadata_path / "metadata.json"
        path.write_text(self._metadata.model_dump_json(indent=4))
        with suppress(shutil.SameFileError):
            self.fs.put_file(path, "metadata/metadata.json")

    @staticmethod
    def _construct_url(
        bucket_storage: BucketStorage,
        bucket: str,
        team_id: str,
        dataset_name: str,
    ) -> str:
        """Constructs a URL for a remote dataset."""
        return f"{bucket_storage.value}://{bucket}/{team_id}/datasets/{dataset_name}"

    # TODO: Is the cache used anywhere at all?
    def _init_credentials(self) -> Dict[str, Any]:
        credentials_cache_file = self.base_path / "credentials.json"
        if credentials_cache_file.exists():
            return json.loads(credentials_cache_file.read_text())
        return {}

    def _get_metadata(self) -> Metadata:
        """Loads metadata from local storage or cloud, depending on the
        BucketStorage type.

        If loads from cloud it always downloads before loading.
        """
        if self.fs.exists("metadata/metadata.json"):
            path = get_file(
                self.fs,
                "metadata/metadata.json",
                self.metadata_path,
                default=self.metadata_path / "metadata.json",
            )
            metadata_json = json.loads(path.read_text())
            version = Version.parse(metadata_json.get("ldf_version", "1.0.0"))
            if version != LDF_VERSION:  # pragma: no cover
                return migrate_metadata(
                    metadata_json,
                    self._load_df_offline(lazy=True, attempt_migration=False),
                )
            return Metadata(**metadata_json)
        return Metadata(
            source=LuxonisSource(),
            ldf_version=str(LDF_VERSION),
            classes={},
            tasks={},
            skeletons={},
            categorical_encodings={},
            metadata_types={},
        )

    @property
    def is_remote(self) -> bool:
        return self.bucket_storage != BucketStorage.LOCAL

    @override
    def update_source(self, source: LuxonisSource) -> None:
        self._metadata.source = source
        self._write_metadata()

    @override
    def set_classes(
        self,
        classes: Union[List[str], Dict[str, int]],
        task: Optional[str] = None,
    ) -> None:
        if task is None:
            tasks = self.get_task_names()
        else:
            tasks = [task]

        for t in tasks:
            self._metadata.set_classes(classes, t)

        self._write_metadata()

    @override
    def get_classes(self) -> Dict[str, Dict[str, int]]:
        return self._metadata.classes

    def get_n_classes(self) -> Dict[str, int]:
        """Returns a mapping of task names to number of classes.

        @rtype: Dict[str, int]
        @return: A mapping from task names to number of classes.
        """
        return {
            task_name: len(classes)
            for task_name, classes in self.get_classes().items()
        }

    @override
    def set_skeletons(
        self,
        labels: Optional[List[str]] = None,
        edges: Optional[List[Tuple[int, int]]] = None,
        task: Optional[str] = None,
    ) -> None:
        if labels is None and edges is None:
            raise ValueError("Must provide either keypoint names or edges")

        if task is None:
            tasks = self.get_task_names()
        else:
            tasks = [task]
        for t in tasks:
            self._metadata.skeletons[t] = {
                "labels": labels or [],
                "edges": sorted(edges or []),
            }
        self._write_metadata()

    @override
    def get_skeletons(
        self,
    ) -> Dict[str, Tuple[List[str], List[Tuple[int, int]]]]:
        return {
            task: (skel["labels"], skel["edges"])
            for task, skel in self._metadata.skeletons.items()
        }

    @override
    def get_tasks(self) -> Dict[str, List[str]]:
        return self._metadata.tasks

    @override
    def set_tasks(self, tasks: Mapping[str, Iterable[str]]) -> None:
        if len(tasks) == 0:
            return
        self._metadata.tasks = {
            task_name: sorted(task_types)
            for task_name, task_types in tasks.items()
        }
        self._write_metadata()

    def get_categorical_encodings(
        self,
    ) -> Dict[str, Dict[str, int]]:
        return self._metadata.categorical_encodings

    def get_metadata_types(
        self,
    ) -> Dict[str, Literal["float", "int", "str", "Category"]]:
        return self._metadata.metadata_types

    def pull_from_cloud(
        self, update_mode: UpdateMode = UpdateMode.MISSING
    ) -> None:
        """Synchronizes the dataset from a remote cloud bucket to the
        local directory.

        This method performs the download only if some local dataset media files are missing, or always downloads
        depending on the provided update_mode.

        @type update_mode: UpdateMode
        @param update_mode: Specifies the update behavior.
            - UpdateMode.MISSING: Downloads only the missing media files for the dataset.
            - UpdateMode.ALL: Always downloads and overwrites all media files in the local dataset.
        """
        if not self.is_remote:
            logger.warning("This is a local dataset! Cannot sync from cloud.")
            return

        local_dir = self.base_path / "data" / self.team_id / "datasets"
        local_dir.mkdir(exist_ok=True, parents=True)

        lock_path = local_dir / ".sync.lock"

        with FileLock(str(lock_path)):  # DDP GCS training - multiple processes
            logger.info(
                "Pulling remote's dataset annotations and metadata to local dataset ..."
            )
            for dir_name in ["annotations", "metadata"]:
                _ = get_dir(self.fs, dir_name, self.local_path)

            index = self._get_index(lazy=False)
            missing_media_paths = []
            if index is not None:
                missing = index.filter(
                    pl.col("original_filepath").apply(
                        lambda path: not Path(path).exists(),
                        return_dtype=pl.Boolean,
                    )
                    & pl.col("uuid").apply(
                        lambda uid: not (
                            Path("local_dir")
                            / "media"
                            / f"{uid}{Path(str(uid)).suffix}"
                        ).exists(),
                        return_dtype=pl.Boolean,
                    )
                )
                missing_media_paths = [
                    f"media/{row[0]}{Path(str(row[1])).suffix}"
                    for row in missing.select(["uuid", "file"]).iter_rows()
                ]

            if update_mode == UpdateMode.ALL:
                logger.info("Force-pulling all media files...")
                self.fs.get_dir(remote_paths="", local_dir=local_dir)
            elif update_mode == UpdateMode.MISSING and missing_media_paths:
                logger.info(
                    f"Pulling {len(missing_media_paths)} missing files..."
                )
                self.fs.get_dir(
                    remote_paths=missing_media_paths,
                    local_dir=local_dir / f"{self.dataset_name}" / "media",
                )
            else:
                logger.info("Media already synced")

    def push_to_cloud(
        self,
        bucket_storage: BucketStorage,
        update_mode: UpdateMode = UpdateMode.MISSING,
    ) -> None:
        """Pushes the local dataset to a remote cloud bucket.

        This method performs the pushing only for missing cloud media files or always pushes
        depending on the provided update_mode.

        @type update_mode: UpdateMode
        @param update_mode: Specifies the update behavior. Annotations and metadata are always pushed.
            - UpdateMode.MISSING: Pushes only the missing media files for the dataset.
            - UpdateMode.ALL:  Always pushes and overwrites all media files in the cloud dataset.
        """
        index = self._get_index(lazy=False)

        if index is None:
            raise ValueError(
                "Cannot push to cloud. The dataset is empty or not initialized."
            )

        dataset = LuxonisDataset(
            dataset_name=self.dataset_name,
            team_id=self.team_id,
            bucket_type=self.bucket_type,
            bucket_storage=bucket_storage,
            delete_local=False,
            delete_remote=False,
        )

        bucket_uuids = (
            [
                PurePosixPath(path).stem
                for path in dataset.fs.walk_dir(
                    "media", recursive=False, typ="file"
                )
            ]
            if dataset.fs.exists("media")
            else []
        )

        missing_df = index.filter(~pl.col("uuid").is_in(bucket_uuids))

        missing_uuid_dict = {}
        for original_path, uuid in zip(
            missing_df["original_filepath"].to_list(),
            missing_df["uuid"].to_list(),
        ):
            if not Path(original_path).exists():
                suffix = Path(original_path).suffix
                fallback_path = self.local_path / "media" / f"{uuid}{suffix}"
                if fallback_path.exists():
                    missing_uuid_dict[str(fallback_path)] = uuid
                else:
                    raise FileNotFoundError(
                        f"File {original_path} and {fallback_path} do not exist!"
                    )
            else:
                missing_uuid_dict[original_path] = uuid

        for dir_name in ["annotations", "metadata"]:
            dataset.fs.put_dir(
                local_paths=self.local_path / dir_name,
                remote_dir=dir_name,
                copy_contents=True,
            )

        if update_mode == UpdateMode.ALL:
            logger.info("Force-pushing all media files...")
            dataset.fs.put_dir(
                local_paths=self.local_path / "media", remote_dir="media"
            )
        elif update_mode == UpdateMode.MISSING and missing_uuid_dict:
            logger.info(
                f"Pushing {len(missing_uuid_dict)} missing files to cloud..."
            )
            dataset.fs.put_dir(
                local_paths=missing_uuid_dict.keys(),
                remote_dir="media",
                uuid_dict=missing_uuid_dict,
            )
        else:
            logger.info("Media already synced")

    @override
    def delete_dataset(
        self, *, delete_remote: bool = False, delete_local: bool = False
    ) -> None:
        """Deletes the dataset from local storage and optionally from
        the cloud.

        @type delete_remote: bool
        @param delete_remote: Whether to delete the dataset from the
            cloud.
        @type delete_local: bool
        @param delete_local: Whether to delete the dataset from local
            storage.
        """
        if not self.is_remote and delete_local:
            logger.info(
                f"Deleting local dataset '{self.dataset_name}' from local storage"
            )
            shutil.rmtree(self.path)

        if self.is_remote and delete_remote:
            logger.info(
                f"Deleting remote dataset '{self.dataset_name}' from cloud storage"
            )
            assert self.path
            assert self.dataset_name
            assert self.local_path
            self.fs.delete_dir(allow_delete_parent=True)

        if self.is_remote and delete_local:
            logger.info(
                f"Deleting remote dataset '{self.dataset_name}' from local storage"
            )
            if self.local_path.exists():
                shutil.rmtree(self.local_path)

    def _process_arrays(self, data_batch: List[DatasetRecord]) -> None:
        logger.info("Checking arrays...")
        task = self.progress.add_task(
            "[magenta]Processing arrays...", total=len(data_batch)
        )
        self.progress.start()
        uuid_dict = {}
        for record in data_batch:
            self.progress.update(task, advance=1)
            if record.annotation is None or record.annotation.array is None:
                continue
            ann = record.annotation.array
            if self.is_remote:
                uuid = self.fs.get_file_uuid(ann.path, local=True)
                uuid_dict[str(ann.path)] = uuid
                ann.path = Path(uuid).with_suffix(ann.path.suffix)
            else:
                ann.path = ann.path.absolute().resolve()
        self.progress.stop()
        self.progress.remove_task(task)
        if self.is_remote:
            logger.info("Uploading arrays...")
            self.fs.put_dir(
                local_paths=uuid_dict.keys(),
                remote_dir="arrays",
                uuid_dict=uuid_dict,
            )

    def _add_process_batch(
        self,
        data_batch: List[DatasetRecord],
        pfm: ParquetFileManager,
        index: Optional[pl.DataFrame],
    ) -> None:
        paths = {data.file for data in data_batch}
        logger.info("Generating UUIDs...")
        uuid_dict = self.fs.get_file_uuids(paths, local=True)

        overwrite_uuids = set()
        for file_path in paths:
            matched_id = find_filepath_uuid(file_path, index)
            if matched_id is not None:
                overwrite_uuids.add(matched_id)
                logger.warning(
                    f"File {file_path} with UUID: {matched_id} already existed in the dataset from previous dataset.add() call. "
                    "Old data will be overwritten with the new data."
                )

        if overwrite_uuids:
            pfm.remove_duplicate_uuids(overwrite_uuids)

        if self.is_remote:
            logger.info("Uploading media...")

            # TODO: support from bucket (likely with a self.fs.copy_dir)
            self.fs.put_dir(
                local_paths=paths,
                remote_dir="media",
                uuid_dict=dict(uuid_dict),
            )
            logger.info("Media uploaded")

        self._process_arrays(data_batch)

        task = self.progress.add_task(
            "[magenta]Processing data...", total=len(data_batch)
        )

        logger.info("Saving annotations...")
        with self.progress:
            for record in data_batch:
                filepath = record.file
                uuid = uuid_dict[str(filepath)]
                for row in record.to_parquet_rows():
                    pfm.write(uuid, row)
                self.progress.update(task, advance=1)
        self.progress.remove_task(task)

    def add(
        self, generator: DatasetIterator, batch_size: int = 1_000_000
    ) -> Self:
        logger.info(f"Adding data to dataset '{self.dataset_name}'...")

        data_batch: list[DatasetRecord] = []

        classes_per_task: Dict[str, Set[str]] = defaultdict(set)
        tasks: Dict[str, Set[str]] = defaultdict(set)
        categorical_encodings = defaultdict(dict)
        metadata_types = {}
        num_kpts_per_task: Dict[str, int] = {}

        annotations_path = get_dir(
            self.fs,
            "annotations",
            self.local_path,
            default=self.annotations_path,
        )

        index = self._get_index()

        assert annotations_path is not None

        with ParquetFileManager(annotations_path, batch_size) as pfm:
            for i, record in enumerate(generator, start=1):
                if not isinstance(record, DatasetRecord):
                    record = DatasetRecord(**record)
                ann = record.annotation
                if ann is not None:
                    if not record.task_name:
                        record.task_name = infer_task(
                            record.task_name,
                            ann.class_name,
                            self.get_classes(),
                        )

                    def update_state(task_name: str, ann: Detection) -> None:
                        if ann.class_name is not None:
                            classes_per_task[task_name].add(ann.class_name)
                        elif not classes_per_task[task_name]:
                            classes_per_task[task_name] = set()

                        tasks[task_name] |= ann.get_task_types()

                        if ann.keypoints is not None:
                            num_kpts_per_task[task_name] = len(
                                ann.keypoints.keypoints
                            )
                        for name, value in ann.metadata.items():
                            task = f"{task_name}/metadata/{name}"
                            typ = type(value).__name__
                            if (
                                task in metadata_types
                                and metadata_types[task] != typ
                            ):
                                if {typ, metadata_types[task]} == {
                                    "int",
                                    "float",
                                }:
                                    metadata_types[task] = "float"
                                else:
                                    raise ValueError(
                                        f"Metadata type mismatch for {task}: {metadata_types[task]} and {typ}"
                                    )
                            else:
                                metadata_types[task] = typ

                            if not isinstance(value, Category):
                                continue
                            if value not in categorical_encodings[task]:
                                categorical_encodings[task][value] = len(
                                    categorical_encodings[task]
                                )
                        for name, sub_detection in ann.sub_detections.items():
                            update_state(f"{task_name}/{name}", sub_detection)

                    update_state(record.task_name, ann)

                data_batch.append(record)
                if i % batch_size == 0:
                    self._add_process_batch(data_batch, pfm, index)
                    data_batch = []

            self._add_process_batch(data_batch, pfm, index)

        with suppress(shutil.SameFileError):
            self.fs.put_dir(annotations_path, "")

        curr_classes = self.get_classes()
        for task, classes in classes_per_task.items():
            old_classes = set(curr_classes.get(task, []))
            new_classes = list(classes - old_classes)
            if new_classes or task not in curr_classes:
                logger.info(
                    f"Detected new classes for task group '{task}': {new_classes}"
                )

                self.set_classes(list(classes | old_classes), task=task)

        for task, num_kpts in num_kpts_per_task.items():
            self.set_skeletons(
                labels=[str(i) for i in range(num_kpts)],
                edges=[(i, i + 1) for i in range(num_kpts - 1)],
                task=task,
            )

        self._metadata.categorical_encodings = dict(categorical_encodings)
        self._metadata.metadata_types = metadata_types
        self.set_tasks(tasks)
        self._warn_on_duplicates()
        return self

    def _warn_on_duplicates(self) -> None:
        df = self._load_df_offline(lazy=True)
        index = self._get_index(lazy=True)
        if df is None or index is None:
            return
        df = df.join(index, on="uuid").drop("file_right")
        warn_on_duplicates(df)

    def get_splits(self) -> Optional[Dict[str, List[str]]]:
        splits_path = get_file(
            self.fs, "metadata/splits.json", self.metadata_path
        )
        if splits_path is None:
            return None

        with open(splits_path) as file:
            return json.load(file)

    @deprecated(
        "ratios",
        "definitions",
        suggest={"ratios": "splits", "definitions": "splits"},
    )
    @override
    def make_splits(
        self,
        splits: Optional[
            Union[
                Mapping[str, Sequence[PathType]],
                Mapping[str, float],
                Tuple[float, float, float],
            ]
        ] = None,
        *,
        ratios: Optional[
            Union[Dict[str, float], Tuple[float, float, float]]
        ] = None,
        definitions: Optional[Dict[str, List[PathType]]] = None,
        replace_old_splits: bool = False,
    ) -> None:
        if ratios is not None and definitions is not None:
            raise ValueError("Cannot provide both ratios and definitions")

        if splits is None and ratios is None and definitions is None:
            splits = {"train": 0.8, "val": 0.1, "test": 0.1}

        if splits is not None:
            if ratios is not None or definitions is not None:
                raise ValueError(
                    "Cannot provide both splits and ratios/definitions"
                )
            if isinstance(splits, tuple):
                ratios = splits
            elif isinstance(splits, dict):
                value = next(iter(splits.values()))
                if isinstance(value, float):
                    ratios = splits  # type: ignore
                elif isinstance(value, list):
                    definitions = splits  # type: ignore

        if ratios is not None:
            if isinstance(ratios, tuple):
                if not len(ratios) == 3:
                    raise ValueError(
                        "Ratios must be a tuple of 3 floats for train, val, and test splits"
                    )
                ratios = {
                    "train": ratios[0],
                    "val": ratios[1],
                    "test": ratios[2],
                }
            sum_ = sum(ratios.values())
            if not math.isclose(sum_, 1.0):
                raise ValueError(f"Ratios must sum to 1.0, got {sum_:0.4f}")

        if definitions is not None:
            n_files = sum(map(len, definitions.values()))
            if n_files > len(self):
                raise ValueError(
                    "Dataset size is smaller than the total number of files in the definitions. "
                    f"Dataset size: {len(self)}, Definitions: {n_files}."
                )

        splits_to_update: List[str] = []
        new_splits: Dict[str, List[str]] = {}
        old_splits: Dict[str, List[str]] = defaultdict(list)

        splits_path = get_file(
            self.fs,
            "metadata/splits.json",
            self.metadata_path,
            default=self.metadata_path / "splits.json",
        )
        if splits_path.exists():
            with open(splits_path) as file:
                old_splits = defaultdict(list, json.load(file))

        defined_uuids = {
            uuid for uuids in old_splits.values() for uuid in uuids
        }

        if definitions is None:
            ratios = ratios or {"train": 0.8, "val": 0.1, "test": 0.1}
            df = self._load_df_offline(raise_when_empty=True)
            ids = (
                df.filter(~pl.col("uuid").is_in(defined_uuids))
                .select("uuid")
                .unique()
                .sort("uuid")
                .get_column("uuid")
                .to_list()
            )
            if not ids:
                if not replace_old_splits:
                    raise ValueError(
                        "No new files to add to splits. "
                        "If you want to generate new splits, set `replace_old_splits=True`"
                    )
                ids = df.select("uuid").unique().get_column("uuid").to_list()
                old_splits = defaultdict(list)

            np.random.shuffle(ids)
            N = len(ids)
            lower_bound = 0
            for split, ratio in ratios.items():
                upper_bound = lower_bound + math.ceil(N * ratio)
                new_splits[split] = ids[lower_bound:upper_bound]
                splits_to_update.append(split)
                lower_bound = upper_bound

        else:
            index = self._get_index()
            if index is None:
                raise FileNotFoundError("File index not found")
            for split, filepaths in definitions.items():
                splits_to_update.append(split)
                if not isinstance(filepaths, list):
                    raise TypeError(
                        "Must provide splits as a list of filepaths"
                    )
                ids = [
                    find_filepath_uuid(filepath, index, raise_on_missing=True)
                    for filepath in filepaths
                ]
                new_splits[split] = ids

        for split, uuids in new_splits.items():
            old_splits[split].extend(uuids)

        splits_path.write_text(json.dumps(old_splits, indent=4))

        with suppress(shutil.SameFileError):
            self.fs.put_file(splits_path, "metadata/splits.json")

    @staticmethod
    @override
    def exists(
        dataset_name: str,
        team_id: Optional[str] = None,
        bucket_storage: BucketStorage = BucketStorage.LOCAL,
        bucket: Optional[str] = None,
    ) -> bool:
        """Checks if a dataset exists.

        @type dataset_name: str
        @param dataset_name: Name of the dataset to check
        @type team_id: Optional[str]
        @param team_id: Optional team identifier
        @type bucket_storage: BucketStorage
        @param bucket_storage: Underlying bucket storage from C{local},
            C{S3}, or C{GCS}. Default is C{local}.
        @type bucket: Optional[str]
        @param bucket: Name of the bucket. Default is C{None}.
        @rtype: bool
        @return: Whether the dataset exists.
        """
        return dataset_name in LuxonisDataset.list_datasets(
            team_id, bucket_storage, bucket
        )

    @staticmethod
    def list_datasets(
        team_id: Optional[str] = None,
        bucket_storage: BucketStorage = BucketStorage.LOCAL,
        bucket: Optional[str] = None,
    ) -> List[str]:
        """Returns a list of all datasets.

        @type team_id: Optional[str]
        @param team_id: Optional team identifier
        @type bucket_storage: BucketStorage
        @param bucket_storage: Underlying bucket storage (local, S3, or
            GCS). Default is local.
        @type bucket: Optional[str]
        @param bucket: Name of the bucket. Default is None.
        @rtype: List[str]
        @return: List of all dataset names.
        """
        base_path = environ.LUXONISML_BASE_PATH
        team_id = team_id or environ.LUXONISML_TEAM_ID

        if bucket_storage == BucketStorage.LOCAL:
            fs = LuxonisFileSystem(
                f"file://{base_path}/data/{team_id}/datasets"
            )
        else:
            bucket = bucket or environ.LUXONISML_BUCKET
            if bucket is None:
                raise ValueError(
                    "Must set LUXONISML_BUCKET environment variable!"
                )
            fs = LuxonisFileSystem(
                LuxonisDataset._construct_url(
                    bucket_storage, bucket, team_id, ""
                )
            )
        if not fs.exists():
            return []

        def process_directory(path: PurePosixPath) -> Optional[str]:
            metadata_path = path / "metadata"
            if fs.exists(metadata_path):
                return path.name
            return None

        paths = (
            PurePosixPath(path)
            for path in fs.walk_dir("", recursive=False, typ="directory")
        )
        with ThreadPoolExecutor() as executor:
            return sorted(
                name for name in executor.map(process_directory, paths) if name
            )

    def export(
        self,
        output_path: PathType,
        dataset_type: DatasetType = DatasetType.NATIVE,
        task_name_to_keep: Optional[str] = None,
        max_partition_size_gb: Optional[float] = None,
        zip_output: bool = False,
    ) -> Union[Path, List[Path]]:
        """Exports the dataset into one of the supported formats.

        @type output_path: PathType
        @param output_path: Path to the directory where the dataset will
            be exported.
        @type dataset_type: DatasetType
        @param dataset_type: To what format to export the dataset.
            Currently only DatasetType.NATIVE is supported.
        @type task_name_to_keep: Optional[str]
        @param task_name_to_keep: Task name to keep. If dataset has
            multiple tasks, this parameter is required.
        @type max_partition_size_gb: Optional[float]
        @param max_partition_size_gb: Maximum size of each partition in
            GB. If the dataset exceeds this size, it will be split into
            multiple partitions named
            {dataset_name}_part{partition_number}. Default is None,
            meaning the dataset will be exported as a single partition
            named {dataset_name}.
        @type zip_output: bool
        @param zip_output: Whether to zip the exported dataset (or each
            partition) after export. Default is False.
        @rtype: Union[Path, List[Path]]
        @return: Path(s) to the ZIP file(s) containing the exported
            dataset.
        """

        def _dump_annotations(
            annotations: Dict[str, List[Any]],
            output_path: Path,
            identifier: str,
            part: Optional[int] = None,
        ) -> None:
            for split_name, annotation_data in annotations.items():
                if part is not None:
                    split_path = (
                        output_path / f"{identifier}_part{part}" / split_name
                    )
                else:
                    split_path = output_path / identifier / split_name
                split_path.mkdir(parents=True, exist_ok=True)
                with open(split_path / "annotations.json", "w") as f:
                    json.dump(annotation_data, f, indent=4)

        if dataset_type is not DatasetType.NATIVE:
            raise NotImplementedError(
                "Only 'NATIVE' dataset export is supported at the moment"
            )
        logger.info(
            f"Exporting '{self.identifier}' to '{dataset_type.name}' format"
        )

        splits = self.get_splits()
        if splits is None:
            raise ValueError("Cannot export dataset without splits")
        if len(self.get_tasks()) > 1 and task_name_to_keep is None:
            raise NotImplementedError(
                "This dataset contains multiple tasks. "
                "Multi-task export is not yet supported; please specify the "
                "'task_name' parameter to export one task at a time."
            )

        if (
            task_name_to_keep is not None
            and task_name_to_keep not in self.get_task_names()
        ):
            raise ValueError(
                f"Task name '{task_name_to_keep}' not found in the dataset. "
                "Please provide a valid task name."
            )

        output_path = Path(output_path)
        if output_path.exists():
            raise ValueError(
                f"Export path '{output_path}' already exists. Please remove it first."
            )
        output_path.mkdir(parents=True)
        image_indices = {}
        annotations = {"train": [], "val": [], "test": []}
        df = self._load_df_offline(raise_when_empty=True)
        if task_name_to_keep is not None:
            df = df.filter(pl.col("task_name").is_in([task_name_to_keep]))
        if not self.is_remote:
            index = self._get_index()
            if index is None:  # pragma: no cover
                raise FileNotFoundError("Cannot find dataset index")
            index = index.filter(pl.col("uuid").is_in(df["uuid"]))
            df = df.join(index, on="uuid").drop("file_right")

        # If instances are not sorted by file, sort them by the first occurrence of the file in the dataset.
        # Within each group, original row ordering is preserved.
        df = (
            df.with_row_count("row_idx")
            .with_columns(
                pl.col("row_idx").min().over("file").alias("first_occur")
            )
            .sort(["first_occur", "row_idx"])
            .drop(["first_occur", "row_idx"])
        )

        splits = self.get_splits()
        assert splits is not None

        current_size = 0
        part = 0 if max_partition_size_gb else None
        max_partition_size = (
            max_partition_size_gb * 1024**3 if max_partition_size_gb else None
        )

        for row in track(
            df.iter_rows(),
            total=len(df),
            description="Exporting ...",
        ):
            uuid = row[7]
            if self.is_remote:
                file_extension = row[0].rsplit(".", 1)[-1]
                file = self.media_path / f"{uuid}.{file_extension}"
                assert file.exists()
            else:
                file = Path(row[-1])

            split = None
            for s, uuids in splits.items():
                if uuid in uuids:
                    split = s
                    break

            assert split is not None

            task_name: str = row[2]
            class_name: Optional[str] = row[3]
            instance_id: int = row[4]
            task_type: str = row[5]
            ann_str: Optional[str] = row[6]

            if file not in image_indices:
                file_size = file.stat().st_size
                if (
                    max_partition_size
                    and current_size + file_size > max_partition_size
                ):
                    _dump_annotations(
                        annotations, output_path, self.identifier, part
                    )
                    current_size = 0
                    assert part is not None
                    part += 1
                    annotations = {"train": [], "val": [], "test": []}

                image_indices[file] = len(image_indices)
                if max_partition_size:
                    data_path = (
                        output_path
                        / f"{self.identifier}_part{part}"
                        / split
                        / "images"
                    )
                else:
                    data_path = (
                        output_path / self.identifier / split / "images"
                    )
                data_path.mkdir(parents=True, exist_ok=True)
                dest_file = data_path / f"{image_indices[file]}{file.suffix}"
                shutil.copy(file, dest_file)
                current_size += file_size

            if ann_str is None:
                annotations[split].append(
                    {
                        "file": str(
                            Path(
                                data_path.name,
                                str(image_indices[file]) + file.suffix,
                            )
                        ),
                        "task_name": task_name,
                    }
                )
                continue

            data = json.loads(ann_str)
            record = {
                "file": str(
                    Path(
                        data_path.name,
                        str(image_indices[file]) + file.suffix,
                    )
                ),
                "task_name": task_name,
                "annotation": {
                    "instance_id": instance_id,
                    "class": class_name,
                },
            }
            if task_type in {
                "instance_segmentation",
                "segmentation",
                "boundingbox",
                "keypoints",
            }:
                record["annotation"][task_type] = data
                annotations[split].append(record)

        _dump_annotations(annotations, output_path, self.identifier, part)

        if zip_output:
            archives = []
            if max_partition_size:
                assert part is not None
                for i in range(part + 1):
                    folder = output_path / f"{self.identifier}_part{i}"
                    if folder.exists():
                        archive_file = shutil.make_archive(
                            str(folder), "zip", root_dir=folder
                        )
                        archives.append(Path(archive_file))
            else:
                folder = output_path / self.identifier
                if folder.exists():
                    archive_file = shutil.make_archive(
                        str(folder), "zip", root_dir=folder
                    )
                    archives.append(Path(archive_file))
            return archives if len(archives) > 1 else archives[0]
        return output_path

    def get_statistics(
        self, sample_size: Optional[int] = None, view: Optional[str] = None
    ) -> Dict[str, Any]:
        """Returns comprehensive dataset statistics as a structured
        dictionary for the given view or the entire dataset.

        The returned dictionary contains:

            - "duplicates": Analysis of duplicated content
                - "duplicate_uuids": List of {"uuid": str, "files": List[str]} for images with same UUID
                - "duplicate_annotations": List of repeated annotations with file_name, task_name, task_type,
                    annotation content, and count

            - "class_distributions": Nested dictionary of class frequencies organized by task_name and task_type
            (excludes classification tasks)

            - "missing_annotations": List of file paths that exist in the dataset but lack annotations

            - "heatmaps": Spatial distribution of annotations as 15x15 grid matrices organized by task_name and task_type

        @type sample_size: Optional[int]
        @param sample_size: Number of samples to use for heatmap generation
        @type view: Optional[str]
        @param view: Name of the view to analyze. If None, the entire dataset is analyzed.
        @rtype: Dict[str, Any]
        @return: Dataset statistics dictionary as described above
        """
        df = self._load_df_offline(lazy=True)
        index = self._get_index(lazy=True)

        stats = {
            "duplicates": {},
            "missing_annotations": 0,
            "heatmaps": {},
        }

        if df is None or index is None:
            return stats

        df = df.join(index, on="uuid").drop("file_right")

        splits = self.get_splits()
        if splits is not None and view and view in splits:
            df = df.filter(pl.col("uuid").is_in(splits[view]))  # type: ignore

        stats["duplicates"] = get_duplicates_info(df)

        stats["class_distributions"] = get_class_distributions(df)

        stats["missing_annotations"] = get_missing_annotations(df)

        stats["heatmaps"] = get_heatmaps(df, sample_size)

        return stats



from pathlib import Path
from typing import List

import pytest
from _pytest.fixtures import SubRequest
from pytest_subtests import SubTests

from luxonis_ml.utils.filesystem import (
    LuxonisFileSystem,
    _get_protocol_and_path,
)

URL_PATH = "luxonis-test-bucket/luxonis-ml-test-data/fs_test_data"

PROTOCOLS = ["file", "gcs", "s3"]


@pytest.fixture
def protocol_tempdir(tempdir: Path, request: SubRequest) -> Path:
    protocol = request.getfixturevalue("protocol")
    tempdir = tempdir / protocol
    tempdir.mkdir(parents=True)
    return tempdir


@pytest.fixture
def local_file(protocol_tempdir: Path, randint: int) -> Path:
    file_path = protocol_tempdir / f"file_{randint}.txt"
    file_path.write_text(f"test {randint}")

    return file_path


@pytest.fixture
def local_dir(protocol_tempdir: Path, randint: int) -> Path:
    dir_path = protocol_tempdir / f"dir_{randint}"
    dir_path.mkdir()

    for i in range(5):
        file_path = dir_path / f"file_{i}.txt"
        file_path.write_text(f"file {randint}")

    (dir_path / "nested").mkdir()
    (dir_path / "nested" / "nested_file.txt").write_text(
        f"nested file {randint}"
    )

    return dir_path


@pytest.fixture(params=PROTOCOLS)
def protocol(request: SubRequest) -> str:
    return request.param


@pytest.fixture
def fs(
    protocol: str,
    python_version: str,
    platform_name: str,
    protocol_tempdir: Path,
) -> LuxonisFileSystem:
    if protocol == "file":
        return LuxonisFileSystem(str(protocol_tempdir))
    url_path = get_platform_specific_url(
        protocol, platform_name, python_version
    )
    return LuxonisFileSystem(url_path)


def test_file(
    fs: LuxonisFileSystem,
    local_file: Path,
    subtests: SubTests,
    protocol_tempdir: Path,
):
    uploaded_file = f"upload_{local_file.name}"
    if fs.exists(uploaded_file):  # pragma: no cover
        fs.delete_file(uploaded_file)

    assert not fs.exists(uploaded_file)

    with subtests.test("upload"):
        fs.put_file(local_file, uploaded_file)
        assert fs.exists(uploaded_file)

    with subtests.test("download"):
        file_path = fs.get_file(
            uploaded_file, protocol_tempdir / f"copy_{uploaded_file}"
        )
        assert file_path.exists()
        assert file_path.read_text() == local_file.read_text()

    with subtests.test("delete"):
        fs.delete_file(uploaded_file)
        assert not fs.exists(uploaded_file)


def test_static_file(
    fs: LuxonisFileSystem,
    local_file: Path,
    protocol_tempdir: Path,
    subtests: SubTests,
    randint: int,
):
    file_name = f"static_file_upload_{randint}.txt"
    url = f"{fs.url}/{file_name}"
    with subtests.test("upload"):
        assert not fs.exists(file_name)
        LuxonisFileSystem.upload(local_file, url)
        assert fs.exists(file_name)

    with subtests.test("download"):
        assert fs.exists(file_name)
        file = LuxonisFileSystem.download(url, protocol_tempdir)
        assert file.exists()
        assert file.read_text() == local_file.read_text()


def test_directory(
    fs: LuxonisFileSystem,
    local_dir: Path,
    subtests: SubTests,
    protocol_tempdir: Path,
):
    with subtests.test("name"):
        uploaded_dir = f"upload_{local_dir.name}_name"
        assert not fs.exists(uploaded_dir)

        fs.put_dir(local_dir, uploaded_dir)
        assert fs.exists(uploaded_dir)
        assert fs.is_directory(uploaded_dir)

        dir_path = fs.get_dir(
            uploaded_dir, protocol_tempdir / f"dowload_{local_dir.name}_name"
        )
        compare_directories(dir_path, local_dir)

    with subtests.test("list"):
        uploaded_dir = f"upload_{local_dir.name}_list"
        assert not fs.exists(uploaded_dir)

        fs.put_dir([str(p) for p in local_dir.rglob("*.txt")], uploaded_dir)

        assert fs.exists(uploaded_dir)
        assert fs.is_directory(uploaded_dir)

        dir_path = fs.get_dir(
            [f"{uploaded_dir}/file_{i}.txt" for i in range(5)]
            + [f"{uploaded_dir}/nested_file.txt"],
            protocol_tempdir / f"dowload_{local_dir.name}_list",
        )
        compare_directories(dir_path, local_dir, flat=True)


def test_static_directory(
    fs: LuxonisFileSystem,
    local_dir: Path,
    protocol_tempdir: Path,
    subtests: SubTests,
    randint: int,
):
    dir_name = f"static_dir_upload_{randint}"
    url = f"{fs.url}/{dir_name}"
    with subtests.test("upload"):
        assert not fs.exists(dir_name)
        LuxonisFileSystem.upload(local_dir, url)
        assert fs.exists(dir_name)

    with subtests.test("download"):
        assert fs.exists(dir_name)
        dir = LuxonisFileSystem.download(url, protocol_tempdir)
        compare_directories(dir, local_dir)


def test_walk_dir(fs: LuxonisFileSystem, local_dir: Path, subtests: SubTests):
    uploaded_dir = f"dir_upload_{local_dir.name}"

    if fs.exists(uploaded_dir):  # pragma: no cover
        fs.delete_dir(uploaded_dir)

    fs.put_dir(local_dir, uploaded_dir)

    def check_walk(expected: List[str], **kwargs) -> None:
        paths = {
            Path(path).relative_to(Path(uploaded_dir))
            for path in fs.walk_dir(uploaded_dir, **kwargs)
        }
        assert paths == set(map(Path, expected))

    with subtests.test("recursive"):
        check_walk(
            [f"file_{i}.txt" for i in range(5)] + ["nested/nested_file.txt"],
            recursive=True,
            typ="file",
        )

    with subtests.test("all"):
        check_walk(
            [f"file_{i}.txt" for i in range(5)] + ["nested"],
            recursive=False,
            typ="all",
        )

    with subtests.test("directories"):
        check_walk(["nested"], recursive=False, typ="directory")


def test_protocol():
    assert _get_protocol_and_path("foo://bar/baz") == ("foo", "bar/baz")
    assert _get_protocol_and_path("gs://foo/bar") == ("gcs", "foo/bar")
    assert _get_protocol_and_path("local_path/to/file") == (
        "file",
        "local_path/to/file",
    )

    with pytest.raises(ValueError, match="Protocol 'foo://' not supported"):
        LuxonisFileSystem("foo://bar")


def test_fail(tempdir: Path, randint: int):
    file = tempdir / f"file_{randint}.txt"
    file.write_text("test")
    with pytest.raises(ValueError, match="Local filesystem is not allowed"):
        LuxonisFileSystem(str(file), allow_local=False)


def test_bytes(fs: LuxonisFileSystem, randint: int):
    bytes_file = f"bytes_test_{randint}.txt"
    fs.put_bytes(f"bytes test {randint}".encode(), bytes_file)
    assert fs.exists(bytes_file)
    buffer = fs.read_to_byte_buffer(bytes_file)
    assert buffer.read() == f"bytes test {randint}".encode()


def compare_directories(
    dir_path: Path, orig_dir_path: Path, flat: bool = False
):
    assert dir_path.exists()
    for i in range(5):
        file_path = dir_path / f"file_{i}.txt"
        assert file_path.exists()
        assert (
            file_path.read_text()
            == (orig_dir_path / f"file_{i}.txt").read_text()
        )
    if flat:
        file_path = "nested_file.txt"
    else:
        file_path = "nested/nested_file.txt"
    nested_file = dir_path / file_path

    assert nested_file.exists()
    assert (
        nested_file.read_text()
        == (orig_dir_path / "nested" / "nested_file.txt").read_text()
    )


def get_platform_specific_url(
    protocol: str, platform: str, python_version: str
):
    return f"{protocol}://{URL_PATH}_{platform}_{python_version}"



import shutil
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, NoReturn

import numpy as np
import pytest
from pytest_subtests.plugin import SubTests

from luxonis_ml.data import (
    BucketStorage,
    Category,
    LuxonisDataset,
    LuxonisLoader,
    LuxonisParser,
    LuxonisSource,
    UpdateMode,
)
from luxonis_ml.data.datasets.base_dataset import DatasetIterator
from luxonis_ml.data.utils.task_utils import get_task_type
from luxonis_ml.enums import DatasetType
from luxonis_ml.typing import Params

from .utils import create_dataset, create_image, get_loader_output


def test_dataset(
    bucket_storage: BucketStorage,
    dataset_name: str,
    augmentation_config: List[Params],
    subtests: SubTests,
    storage_url: str,
    tempdir: Path,
):
    with subtests.test("test_create"):
        dataset = LuxonisParser(
            f"{storage_url}/COCO_people_subset.zip",
            dataset_name=dataset_name,
            save_dir=tempdir,
            dataset_type=DatasetType.COCO,
            bucket_storage=bucket_storage,
            delete_local=True,
            delete_remote=True,
            task_name="coco",
        ).parse()

        assert LuxonisDataset.exists(
            dataset_name, bucket_storage=bucket_storage
        )
        assert set(dataset.get_task_names()) == {"coco"}
        assert dataset.get_classes().get("coco") == {"person": 0}
        assert dataset.get_skeletons() == {
            "coco": (
                [
                    "nose",
                    "left_eye",
                    "right_eye",
                    "left_ear",
                    "right_ear",
                    "left_shoulder",
                    "right_shoulder",
                    "left_elbow",
                    "right_elbow",
                    "left_wrist",
                    "right_wrist",
                    "left_hip",
                    "right_hip",
                    "left_knee",
                    "right_knee",
                    "left_ankle",
                    "right_ankle",
                ],
                sorted(
                    [
                        (15, 13),
                        (13, 11),
                        (16, 14),
                        (14, 12),
                        (11, 12),
                        (5, 11),
                        (6, 12),
                        (5, 6),
                        (5, 7),
                        (6, 8),
                        (7, 9),
                        (8, 10),
                        (1, 2),
                        (0, 1),
                        (0, 2),
                        (1, 3),
                        (2, 4),
                        (3, 5),
                        (4, 6),
                    ]
                ),
            ),
        }
        assert dataset.get_n_keypoints() == {"coco": 17}

        assert dataset.identifier == dataset_name

    if "dataset" not in locals():
        pytest.exit("Dataset creation failed")

    with subtests.test("test_source"):
        assert dataset.source == LuxonisSource()
        dataset.update_source(LuxonisSource(name="test"))
        assert dataset.source == LuxonisSource(name="test")

    with subtests.test("test_load"):
        loader = LuxonisLoader(dataset)
        for img, labels in loader:
            assert img is not None
            for task in ["segmentation", "keypoints", "boundingbox"]:
                assert f"coco/{task}" in labels

    with subtests.test("test_load_aug"):
        loader = LuxonisLoader(
            dataset,
            width=512,
            height=512,
            augmentation_config=augmentation_config,
            augmentation_engine="albumentations",
        )
        for img, labels in loader:
            assert img is not None
            for task in ["segmentation", "keypoints", "boundingbox"]:
                assert f"coco/{task}" in labels

    with subtests.test("test_delete"):
        dataset.delete_dataset(delete_remote=True, delete_local=True)
        assert not LuxonisDataset.exists(
            dataset_name, bucket_storage=bucket_storage
        )


@pytest.mark.dependency(name="test_dataset[BucketStorage.LOCAL]")
def test_dataset_fail(dataset_name: str, tempdir: Path):
    def generator() -> DatasetIterator:
        for i in range(10):
            img = create_image(i, tempdir)
            yield {
                "file": img,
                "annotation": {
                    "class": "person",
                },
            }

    dataset = create_dataset(dataset_name, generator())

    with pytest.raises(ValueError, match="Must provide either"):
        dataset.set_skeletons()


@pytest.mark.dependency(name="test_dataset[BucketStorage.LOCAL]")
def test_loader_iterator(storage_url: str, tempdir: Path):
    dataset = LuxonisParser(
        f"{storage_url}/COCO_people_subset.zip",
        dataset_name="_iterator_test",
        save_dir=tempdir,
        dataset_type=DatasetType.COCO,
        delete_local=True,
        task_name="coco",
    ).parse()
    loader = LuxonisLoader(dataset)

    def _raise(*_) -> NoReturn:
        raise IndexError

    loader._load_data = _raise  # type: ignore
    with pytest.raises(IndexError):
        _ = loader[0]


@pytest.mark.dependency(name="test_dataset[BucketStorage.LOCAL]")
def test_make_splits(
    bucket_storage: BucketStorage, dataset_name: str, tempdir: Path
):
    definitions: Dict[str, List[str]] = defaultdict(list)

    _start_index: int = 0

    def generator(step: int = 15) -> DatasetIterator:
        nonlocal _start_index
        definitions.clear()
        for i in range(_start_index, _start_index + step):
            path = create_image(i, tempdir)
            yield {
                "file": str(path),
                "annotation": {
                    "class": ["dog", "cat"][i % 2],
                },
            }
            definitions[["train", "val", "test"][i % 3]].append(str(path))
        _start_index += step

    dataset = create_dataset(
        dataset_name, generator(), bucket_storage, splits=False
    )

    assert len(dataset) == 15
    assert dataset.get_splits() is None
    dataset.make_splits(definitions)
    splits = dataset.get_splits()
    assert splits is not None
    assert set(splits.keys()) == {"train", "val", "test"}
    for split, split_data in splits.items():
        assert len(split_data) == 5, (
            f"Split {split} has {len(split_data)} samples"
        )

    dataset.add(generator())
    splits = dataset.get_splits()
    assert splits is not None
    for split, split_data in splits.items():
        assert len(split_data) == 5, (
            f"Split {split} has {len(split_data)} samples"
        )
    dataset.make_splits(definitions)  # type: ignore
    splits = dataset.get_splits()
    assert splits is not None
    for split, split_data in splits.items():
        assert len(split_data) == 10, (
            f"Split {split} has {len(split_data)} samples"
        )

    dataset.add(generator())
    dataset.make_splits((1, 0, 0))
    splits = dataset.get_splits()
    assert splits is not None
    for split, split_data in splits.items():
        expected_length = 25 if split == "train" else 10
        assert len(split_data) == expected_length, (
            f"Split {split} has {len(split_data)} samples"
        )

    with pytest.raises(ValueError, match="No new files"):
        dataset.make_splits()

    with pytest.raises(ValueError, match="Ratios must sum to 1.0"):
        dataset.make_splits((0.7, 0.1, 1))

    with pytest.raises(ValueError, match="must be a tuple of 3 floats"):
        dataset.make_splits((0.7, 0.1, 0.1, 0.1))  # type: ignore

    with pytest.raises(ValueError, match="Cannot provide both splits and"):
        dataset.make_splits((0.7, 0.1, 0.2), definitions=definitions)

    with pytest.raises(ValueError, match="Ratios must sum to 1.0"):
        dataset.make_splits({"train": 1.5})

    with pytest.raises(ValueError, match="Dataset size is smaller than"):
        dataset.make_splits(
            {split: defs * 2 for split, defs in splits.items()}
        )

    dataset.add(generator(10))
    dataset.make_splits({"custom_split": 1.0})
    splits = dataset.get_splits()
    assert splits is not None
    assert set(splits.keys()) == {"train", "val", "test", "custom_split"}
    for split, split_data in splits.items():
        expected_length = 25 if split == "train" else 10
        assert len(split_data) == expected_length, (
            f"Split {split} has {len(split_data)} samples"
        )

    dataset.make_splits(replace_old_splits=True)
    splits = dataset.get_splits()
    assert splits is not None
    for split, split_data in splits.items():
        expected_length = {"train": 44, "val": 6, "test": 5}
        assert len(split_data) == expected_length[split], (
            f"Split {split} has {len(split_data)} samples"
        )


@pytest.mark.dependency(name="test_dataset[BucketStorage.LOCAL]")
def test_metadata(
    bucket_storage: BucketStorage, dataset_name: str, tempdir: Path
):
    def generator() -> DatasetIterator:
        img = create_image(0, tempdir)
        for i in range(10):
            yield {
                "file": img,
                "annotation": {
                    "class": "person",
                    "metadata": {
                        "color": Category("red" if i % 2 == 0 else "blue"),
                        "distance": 5.0 if i == 0 else 5,
                        "id": 127 + i,
                        "license_plate": "xyz",
                    },
                },
            }

    dataset = create_dataset(dataset_name, generator(), bucket_storage)
    loader = LuxonisLoader(dataset)
    for _, labels in loader:
        labels = {get_task_type(k): v for k, v in labels.items()}
        assert {
            "metadata/color",
            "metadata/distance",
            "metadata/id",
            "metadata/license_plate",
            "classification",
        } == set(labels.keys())

        assert labels["metadata/color"].tolist() == [0, 1] * 5
        assert labels["metadata/distance"].tolist() == [5.0] * 10
        assert labels["metadata/id"].tolist() == list(range(127, 137))
        assert labels["metadata/license_plate"].tolist() == ["xyz"] * 10

    loader = LuxonisLoader(dataset, keep_categorical_as_strings=True)
    for _, labels in loader:
        labels = {get_task_type(k): v for k, v in labels.items()}
        assert labels["metadata/color"].tolist() == ["red", "blue"] * 5

    assert dataset.get_categorical_encodings() == {
        "/metadata/color": {"red": 0, "blue": 1}
    }

    assert dataset.get_metadata_types() == {
        "/metadata/color": "Category",
        "/metadata/distance": "float",
        "/metadata/id": "int",
        "/metadata/license_plate": "str",
    }


@pytest.mark.dependency(name="test_dataset[BucketStorage.LOCAL]")
def test_no_labels(dataset_name: str, tempdir: Path, subtests: SubTests):
    def generator(total: bool) -> DatasetIterator:
        for i in range(10):
            img = create_image(i, tempdir)
            if i == 0:
                if total:
                    yield {
                        "file": img,
                    }
                else:
                    yield {
                        "file": img,
                        "annotation": {
                            "class": "person",
                            "boundingbox": {
                                "x": 0.1,
                                "y": 0.1,
                                "w": 0.1,
                                "h": 0.1,
                            },
                            "keypoints": {
                                "keypoints": [[0.1, 0.1, 0], [0.2, 0.2, 1]]
                            },
                            "segmentation": {
                                "mask": np.random.rand(512, 512) > 0.5
                            },
                            "instance_segmentation": {
                                "mask": np.random.rand(512, 512) > 0.5
                            },
                            "metadata": {
                                "color": "red",
                            },
                            "sub_detections": {
                                "head": {
                                    "boundingbox": {
                                        "x": 0.2,
                                        "y": 0.2,
                                        "w": 0.1,
                                        "h": 0.1,
                                    },
                                },
                            },
                        },
                    }

    for total in [True, False]:
        with subtests.test(f"test_{'total' if total else 'almost'}_empty"):
            dataset = create_dataset(dataset_name, generator(total=total))

            if total:
                expected_tasks = set()
            else:
                expected_tasks = {
                    "/classification",
                    "/boundingbox",
                    "/keypoints",
                    "/segmentation",
                    "/instance_segmentation",
                    "/metadata/color",
                    "/head/boundingbox",
                }

            for _, labels in LuxonisLoader(dataset):
                assert set(labels.keys()) == expected_tasks

            augmented_loader = LuxonisLoader(
                dataset,
                width=512,
                height=512,
                augmentation_config=[{"name": "Flip", "params": {}}],
            )
            for _, labels in augmented_loader:
                assert set(labels.keys()) == expected_tasks

        if total is False:
            with subtests.test("test_almost_empty_exclude_empty"):
                for i, (_, labels) in enumerate(
                    LuxonisLoader(dataset, exclude_empty_annotations=True)
                ):
                    if i == 0:
                        assert set(labels.keys()) == expected_tasks
                    else:
                        assert not labels


@pytest.mark.dependency(name="test_dataset[BucketStorage.LOCAL]")
def test_deep_nested_labels(
    dataset_name: str, augmentation_config: List[Params], tempdir: Path
):
    def generator() -> DatasetIterator:
        for i in range(10):
            yield {
                "file": create_image(i, tempdir),
                "annotation": {
                    "class": "car",
                    "boundingbox": {
                        "x": 0.1,
                        "y": 0.1,
                        "w": 0.1,
                        "h": 0.1,
                    },
                    "sub_detections": {
                        "license_plate": {
                            "class": "CO",
                            "boundingbox": {
                                "x": 0.2,
                                "y": 0.2,
                                "w": 0.1,
                                "h": 0.1,
                            },
                            "metadata": {
                                "text": "ABC123",
                            },
                            "sub_detections": {
                                "text": {
                                    "boundingbox": {
                                        "x": 0.3,
                                        "y": 0.3,
                                        "w": 0.1,
                                        "h": 0.1,
                                    }
                                },
                            },
                        },
                        "driver": {
                            "boundingbox": {
                                "x": 0.3,
                                "y": 0.3,
                                "w": 0.1,
                                "h": 0.1,
                            },
                            "keypoints": {
                                "keypoints": [(0.1, 0.1, 2), (0.2, 0.2, 2)]
                            },
                        },
                    },
                },
            }

    dataset = create_dataset(dataset_name, generator())
    assert dataset.get_tasks() == {
        "": ["boundingbox", "classification"],
        "/driver": ["boundingbox", "keypoints"],
        "/license_plate": [
            "boundingbox",
            "classification",
            "metadata/text",
        ],
        "/license_plate/text": ["boundingbox"],
    }
    assert dataset.get_classes() == {
        "": {"car": 0},
        "/driver": {},
        "/license_plate": {"CO": 0},
        "/license_plate/text": {},
    }
    loader = LuxonisLoader(
        dataset,
        height=512,
        width=512,
        augmentation_config=augmentation_config,
        augmentation_engine="albumentations",
    )
    assert get_loader_output(loader) == {
        "/classification",
        "/boundingbox",
        "/driver/boundingbox",
        "/driver/keypoints",
        "/license_plate/boundingbox",
        "/license_plate/classification",
        "/license_plate/metadata/text",
        "/license_plate/text/boundingbox",
    }


@pytest.mark.dependency(name="test_dataset[BucketStorage.LOCAL]")
def test_partial_labels(dataset_name: str, tempdir: Path):
    def generator() -> DatasetIterator:
        for i in range(8):
            img = create_image(i, tempdir)
            if i < 2:
                yield {
                    "file": img,
                }
            elif i < 4:
                yield {
                    "file": img,
                    "annotation": {
                        "class": "dog",
                    },
                }
            elif i < 6:
                yield {
                    "file": img,
                    "annotation": {
                        "class": "dog",
                        "boundingbox": {
                            "x": 0.1,
                            "y": 0.1,
                            "w": 0.1,
                            "h": 0.1,
                        },
                        "keypoints": {
                            "keypoints": [[0.1, 0.1, 0], [0.2, 0.2, 1]]
                        },
                    },
                }
            elif i < 8:
                yield {
                    "file": img,
                    "annotation": {
                        "class": "dog",
                        "segmentation": {
                            "mask": np.random.rand(512, 512) > 0.5
                        },
                    },
                }

    dataset = create_dataset(dataset_name, generator())

    loader = LuxonisLoader(
        dataset,
        width=512,
        height=512,
        augmentation_config=[{"name": "Rotate"}],
    )

    assert get_loader_output(loader) == {
        "/classification",
        "/boundingbox",
        "/keypoints",
        "/segmentation",
    }


@pytest.mark.dependency(name="test_dataset[BucketStorage.LOCAL]")
def test_clone_dataset(
    bucket_storage: BucketStorage, dataset_name: str, tempdir: Path
):
    def generator1() -> DatasetIterator:
        for i in range(3):
            img = create_image(i, tempdir)
            yield {
                "file": img,
                "annotation": {
                    "class": "person",
                    "boundingbox": {"x": 0.1, "y": 0.1, "w": 0.1, "h": 0.1},
                },
            }

    dataset = create_dataset(
        dataset_name,
        generator1(),
        bucket_storage,
        splits={"train": 0.6, "val": 0.4},
    )

    cloned_dataset = dataset.clone(new_dataset_name=dataset_name + "_cloned")
    if bucket_storage == BucketStorage.GCS:  # test GCS push/pull
        local_stats = cloned_dataset.get_statistics()
        cloned_dataset = LuxonisDataset(
            cloned_dataset.dataset_name,
            bucket_storage=bucket_storage,
            delete_local=True,
            delete_remote=False,
        )
        cloned_dataset.pull_from_cloud(update_mode=UpdateMode.MISSING)
        synced_stats = cloned_dataset.get_statistics()
        assert local_stats == synced_stats

    assert cloned_dataset.get_splits() == dataset.get_splits()
    assert cloned_dataset.get_classes() == dataset.get_classes()
    assert cloned_dataset.get_task_names() == dataset.get_task_names()
    assert cloned_dataset.get_skeletons() == dataset.get_skeletons()

    df_cloned = cloned_dataset._load_df_offline()
    df_original = dataset._load_df_offline()
    assert df_cloned is not None
    assert df_original is not None
    assert df_cloned.equals(df_original)


@pytest.mark.dependency(name="test_dataset[BucketStorage.LOCAL]")
def test_merge_datasets(
    bucket_storage: BucketStorage,
    dataset_name: str,
    tempdir: Path,
    subtests: SubTests,
):
    dataset_name = f"{dataset_name}_{bucket_storage.value}"
    dataset1_name = f"{dataset_name}_1"
    dataset2_name = f"{dataset_name}_2"

    def generator1() -> DatasetIterator:
        for i in range(3):
            img = create_image(i, tempdir)
            yield {
                "file": img,
                "annotation": {
                    "class": "person",
                    "boundingbox": {"x": 0.1, "y": 0.1, "w": 0.1, "h": 0.1},
                },
            }

    def generator2() -> DatasetIterator:
        for i in range(3, 6):
            img = create_image(i, tempdir)
            yield {
                "file": img,
                "annotation": {
                    "class": "dog",
                    "boundingbox": {"x": 0.2, "y": 0.2, "w": 0.2, "h": 0.2},
                },
            }

    dataset1 = create_dataset(
        dataset1_name,
        generator1(),
        bucket_storage,
        splits={"train": 0.6, "val": 0.4},
    )

    dataset2 = create_dataset(
        dataset2_name,
        generator2(),
        bucket_storage,
        splits={"train": 0.6, "val": 0.4},
    )

    with subtests.test("test_inplace"):
        cloned_dataset1 = dataset1.clone(
            new_dataset_name=f"{dataset1_name}_cloned"
        )
        cloned_dataset1_merged_with_dataset2 = cloned_dataset1.merge_with(
            dataset2, inplace=True
        )

        if bucket_storage == BucketStorage.GCS:  # test GCS push/pull
            local_stats = cloned_dataset1_merged_with_dataset2.get_statistics()
            cloned_dataset1_merged_with_dataset2 = LuxonisDataset(
                cloned_dataset1_merged_with_dataset2.dataset_name,
                bucket_storage=bucket_storage,
                delete_local=True,
                delete_remote=False,
            )
            cloned_dataset1_merged_with_dataset2.pull_from_cloud(
                update_mode=UpdateMode.MISSING
            )
            synced_stats = (
                cloned_dataset1_merged_with_dataset2.get_statistics()
            )
            synced_stats["class_distributions"][""]["boundingbox"] = sorted(
                synced_stats["class_distributions"][""]["boundingbox"],
                key=lambda x: x["class_name"],
            )
            local_stats["class_distributions"][""]["boundingbox"] = sorted(
                local_stats["class_distributions"][""]["boundingbox"],
                key=lambda x: x["class_name"],
            )
            assert local_stats == synced_stats

        cloned_dataset1_merged_with_dataset2_stats = (
            cloned_dataset1_merged_with_dataset2.get_statistics()
        )
        found = {
            (item["count"], item["class_name"])
            for item in cloned_dataset1_merged_with_dataset2_stats[
                "class_distributions"
            ][""]["boundingbox"]
        }
        assert found == {(3, "person"), (3, "dog")}

        classes = cloned_dataset1_merged_with_dataset2.get_classes()
        assert set(classes[""]) == {"person", "dog"}

    with subtests.test("test_out_of_place"):
        dataset1_merged_with_dataset2 = dataset1.merge_with(
            dataset2,
            inplace=False,
            new_dataset_name=f"{dataset1_name}_{dataset2_name}_merged",
        )

        if bucket_storage == BucketStorage.GCS:  # test GCS push/pull
            local_stats = dataset1_merged_with_dataset2.get_statistics()
            dataset1_merged_with_dataset2 = LuxonisDataset(
                dataset1_merged_with_dataset2.dataset_name,
                bucket_storage=bucket_storage,
                delete_local=True,
                delete_remote=False,
            )
            dataset1_merged_with_dataset2.pull_from_cloud(
                update_mode=UpdateMode.MISSING
            )
            synced_stats = dataset1_merged_with_dataset2.get_statistics()
            synced_stats["class_distributions"][""]["boundingbox"] = sorted(
                synced_stats["class_distributions"][""]["boundingbox"],
                key=lambda x: x["class_name"],
            )
            local_stats["class_distributions"][""]["boundingbox"] = sorted(
                local_stats["class_distributions"][""]["boundingbox"],
                key=lambda x: x["class_name"],
            )
            assert local_stats == synced_stats

    classes = dataset1_merged_with_dataset2.get_classes()
    assert set(classes[""]) == {"person", "dog"}
    dataset1_merged_with_dataset2_stats = (
        dataset1_merged_with_dataset2.get_statistics()
    )
    found = {
        (item["count"], item["class_name"])
        for item in dataset1_merged_with_dataset2_stats["class_distributions"][
            ""
        ]["boundingbox"]
    }
    assert found == {(3, "person"), (3, "dog")}

    df_merged = dataset1_merged_with_dataset2._load_df_offline()
    df_cloned_merged = dataset1.merge_with(
        dataset2, inplace=True
    )._load_df_offline()
    assert df_merged is not None
    assert df_cloned_merged is not None
    assert df_merged.equals(df_cloned_merged)


@pytest.mark.dependency(name="test_dataset[BucketStorage.LOCAL]")
def test_classes_per_task(dataset_name: str, tempdir: Path):
    def generator() -> DatasetIterator:
        img = create_image(0, tempdir)
        yield {
            "file": img,
            "annotation": {
                "class": "person",
                "boundingbox": {"x": 0.1, "y": 0.1, "w": 0.1, "h": 0.1},
                "instance_id": 0,
            },
        }
        # Yield a second annotation with only an `instance_id` to check that we don't encounter the issue: "Detected new classes for task group '': []".
        yield {
            "file": img,
            "annotation": {
                "keypoints": {"keypoints": [[0.1, 0.1, 0], [0.2, 0.2, 1]]},
                "instance_id": 0,
            },
        }

    dataset = create_dataset(dataset_name, generator())

    assert dataset.get_classes() == {"": {"person": 0}}


@pytest.mark.dependency(name="test_dataset[BucketStorage.LOCAL]")
def test_keypoints_solo(dataset_name: str, tempdir: Path):
    def generator() -> DatasetIterator:
        for i in range(4):
            img = create_image(i, tempdir)
            yield {
                "file": img,
                "annotation": {
                    "class": "person",
                    "keypoints": {"keypoints": [[0.1, 0.1, 0], [0.2, 0.2, 1]]},
                },
            }

    augs = [
        {"name": "Normalize"},
        {"name": "Defocus", "params": {"p": 1}},
        {
            "name": "Mosaic4",
            "params": {"out_width": 512, "out_height": 512, "p": 1},
        },
    ]
    dataset = create_dataset(dataset_name, generator())
    loader = LuxonisLoader(
        dataset, height=512, width=512, augmentation_config=augs
    )
    for _ in loader:
        pass


@pytest.mark.dependency(name="test_dataset[BucketStorage.LOCAL]")
def test_dataset_push_pull(
    dataset_name: str, tempdir: Path, subtests: SubTests
):
    def generator(start: int, end: int) -> DatasetIterator:
        """Generate sample dataset items with bounding boxes."""
        for i in range(start, end):
            img = create_image(i, tempdir)
            yield {
                "file": img,
                "annotation": {
                    "class": "person",
                    "boundingbox": {"x": 0.1, "y": 0.1, "w": 0.1, "h": 0.1},
                    "instance_id": i,
                },
            }

    with subtests.test("create_initial_dataset"):
        original_dataset = create_dataset(
            dataset_name,
            generator(0, 3),
            bucket_storage=BucketStorage.LOCAL,
            delete_local=True,
            delete_remote=True,
            splits=(1, 0, 0),
        )
        original_stats = original_dataset.get_statistics()

    with subtests.test("verify_dataset_overwrite"):
        overwritten_dataset = create_dataset(
            dataset_name,
            generator(0, 3),
            bucket_storage=BucketStorage.LOCAL,
            delete_local=False,
            delete_remote=False,
            splits=False,
        )
        assert overwritten_dataset.get_statistics() == original_stats

    with subtests.test("push_to_cloud"):
        overwritten_dataset.push_to_cloud(bucket_storage=BucketStorage.GCS)

        overwritten_dataset.delete_dataset(
            delete_local=True, delete_remote=False
        )
        del overwritten_dataset

        assert not LuxonisDataset.exists(
            dataset_name, bucket_storage=BucketStorage.LOCAL
        )
        assert LuxonisDataset.exists(
            dataset_name, bucket_storage=BucketStorage.GCS
        )

    with subtests.test("pull_from_cloud_local_media_empty"):
        cloud_dataset = LuxonisDataset(
            dataset_name,
            bucket_storage=BucketStorage.GCS,
            delete_local=True,
            delete_remote=False,
        )

        cloud_dataset.pull_from_cloud(update_mode=UpdateMode.MISSING)

        assert cloud_dataset.get_statistics() == original_stats
        assert sum(1 for _ in LuxonisLoader(cloud_dataset)) == 3

    with subtests.test("pull_from_cloud_local_media_full"):
        cloud_dataset.delete_dataset(delete_local=True, delete_remote=False)
        shutil.rmtree(tempdir)
        del cloud_dataset

        cloud_dataset_again = LuxonisDataset(
            dataset_name,
            bucket_storage=BucketStorage.GCS,
            delete_local=True,
            delete_remote=False,
        )

        cloud_dataset_again.pull_from_cloud(update_mode=UpdateMode.MISSING)

        assert cloud_dataset_again.get_statistics() == original_stats
        assert sum(1 for _ in LuxonisLoader(cloud_dataset_again)) == 3

        cloud_dataset_again.delete_dataset(
            delete_local=False, delete_remote=True
        )

    with subtests.test("push_to_cloud_local_media_full_loader_resync"):
        local_dataset = LuxonisDataset(
            dataset_name,
            bucket_storage=BucketStorage.LOCAL,
            delete_local=False,
        )
        local_dataset.push_to_cloud(bucket_storage=BucketStorage.GCS)
        assert cloud_dataset_again.get_statistics() == original_stats

        cloud_dataset = LuxonisDataset(
            dataset_name,
            bucket_storage=BucketStorage.GCS,
            delete_local=True,
            delete_remote=False,
        )

        loader = LuxonisLoader(cloud_dataset)
        assert sum(1 for _ in loader) == 3
        assert cloud_dataset.get_statistics() == original_stats



from copy import deepcopy
from typing import Dict

import numpy as np
import pytest

from luxonis_ml.data import AlbumentationsEngine
from luxonis_ml.typing import Labels


@pytest.fixture
def image() -> np.ndarray:
    return np.zeros((320, 320, 3), dtype=np.uint8)


@pytest.fixture
def labels() -> Labels:
    return {
        "task/classification": np.array([1.0]),
        "task/boundingbox": np.array(
            [
                [0.0, 0.57, 0.30, 0.17, 0.25],
                [0.0, 0.39, 0.27, 0.20, 0.10],
            ]
        ),
        "task/keypoints": np.array(
            [
                [0.69, 0.37, 2.0, 0.0, 0.0, 0.0, 0.68, 0.36, 2.0],
                [0.51, 0.33, 2.0, 0.0, 0.0, 0.0, 0.50, 0.32, 2.0],
            ]
        ),
        "task/segmentation": np.zeros((1, 320, 320)),
    }


@pytest.fixture
def targets() -> Dict[str, str]:
    return {
        "task/boundingbox": "boundingbox",
        "task/keypoints": "keypoints",
        "task/segmentation": "segmentation",
    }


@pytest.fixture
def n_classes() -> Dict[str, int]:
    return {
        "task/boundingbox": 1,
        "task/keypoints": 1,
        "task/segmentation": 1,
    }


def test_mosaic4(
    image: np.ndarray,
    labels: Labels,
    targets: Dict[str, str],
    n_classes: Dict[str, int],
):
    config = [
        {
            "name": "Mosaic4",
            "params": {"p": 1.0, "out_width": 640, "out_height": 640},
        }
    ]
    augmentations = AlbumentationsEngine(256, 256, targets, n_classes, config)
    augmentations.apply([(image.copy(), deepcopy(labels)) for _ in range(4)])


def test_mixup(
    image: np.ndarray,
    labels: Labels,
    targets: Dict[str, str],
    n_classes: Dict[str, int],
):
    config = [{"name": "MixUp", "params": {"p": 1.0}}]
    augmentations = AlbumentationsEngine(256, 256, targets, n_classes, config)
    augmentations.apply([(image.copy(), deepcopy(labels)) for _ in range(2)])


def test_batched_p_0(
    image: np.ndarray,
    labels: Labels,
    targets: Dict[str, str],
    n_classes: Dict[str, int],
):
    config = [
        {
            "name": "Mosaic4",
            "params": {"p": 0, "out_width": 640, "out_height": 640},
        },
        {"name": "MixUp", "params": {"p": 0}},
    ]
    augmentations = AlbumentationsEngine(256, 256, targets, n_classes, config)
    augmentations.apply([(image.copy(), deepcopy(labels)) for _ in range(8)])



from typing import Type

import albumentations as A

from luxonis_ml.utils import Registry

from .letterbox_resize import LetterboxResize
from .mixup import MixUp
from .mosaic import Mosaic4

TRANSFORMATIONS: Registry[Type[A.BasicTransform]] = Registry(
    "albumentations_transformations"
)

TRANSFORMATIONS.register(module=LetterboxResize)
TRANSFORMATIONS.register(module=MixUp)
TRANSFORMATIONS.register(module=Mosaic4)

__all__ = ["TRANSFORMATIONS", "LetterboxResize", "MixUp", "Mosaic4"]



from typing import Dict, List

import albumentations as A
import numpy as np
from albumentations.core.composition import TransformsSeqType
from typing_extensions import override

from .batch_transform import BatchTransform
from .utils import yield_batches


class BatchCompose(A.Compose):
    transforms: List[BatchTransform]

    def __init__(self, transforms: TransformsSeqType, **kwargs):
        """Compose transforms and handle all transformations regarding
        bounding boxes.

        @param transforms: List of transformations to compose
        @type transforms: TransformsSeqType
        @param kwargs: Additional arguments to pass to A.Compose
        """
        super().__init__(transforms, is_check_shapes=False, **kwargs)

        self.batch_size = 1
        for transform in self.transforms:
            self.batch_size *= transform.batch_size

    @override
    def __call__(
        self, data_batch: List[Dict[str, np.ndarray]]
    ) -> Dict[str, np.ndarray]:
        if len(data_batch) != self.batch_size:
            raise ValueError(
                f"Batch size must be equal to {self.batch_size}, "
                f"but got {len(data_batch)}."
            )

        if not self.transforms:
            return data_batch[0]

        for data in data_batch:
            self.preprocess(data)

        for transform in self.transforms:
            new_batch = []
            for batch in yield_batches(data_batch, transform.batch_size):
                data = transform(**batch)  # type: ignore

                if isinstance(next(iter(data.values())), list):
                    data = {key: value[0] for key, value in batch.items()}

                data = self.check_data_post_transform(data)
                new_batch.append(data)
            data_batch = new_batch

        assert len(data_batch) == 1
        data = data_batch[0]

        data = self.make_contiguous(data)

        return self.postprocess(data)

    @staticmethod
    def make_contiguous(data: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        for key, value in data.items():
            if isinstance(value, np.ndarray):
                value = np.ascontiguousarray(value)
            data[key] = value
        return data



from .data_utils import (
    find_duplicates,
    get_class_distributions,
    get_duplicates_info,
    get_heatmaps,
    get_missing_annotations,
    infer_task,
    rgb_to_bool_masks,
    warn_on_duplicates,
)
from .enums import BucketStorage, BucketType, ImageType, MediaType, UpdateMode
from .parquet import ParquetFileManager, ParquetRecord
from .plot_utils import plot_class_distribution, plot_heatmap
from .task_utils import (
    get_task_name,
    get_task_type,
    split_task,
    task_is_metadata,
    task_type_iterator,
)
from .visualizations import (
    ColorMap,
    concat_images,
    create_text_image,
    distinct_color_generator,
    visualize,
)

__all__ = [
    "BucketStorage",
    "BucketType",
    "ColorMap",
    "ImageType",
    "MediaType",
    "ParquetFileManager",
    "ParquetRecord",
    "UpdateMode",
    "concat_images",
    "create_text_image",
    "distinct_color_generator",
    "find_duplicates",
    "get_class_distributions",
    "get_duplicates_info",
    "get_heatmaps",
    "get_missing_annotations",
    "get_task_name",
    "get_task_type",
    "infer_task",
    "plot_class_distribution",
    "plot_heatmap",
    "rgb_to_bool_masks",
    "split_task",
    "task_is_metadata",
    "task_type_iterator",
    "visualize",
    "warn_on_duplicates",
]



from luxonis_ml.guard_extras import guard_missing_extra

with guard_missing_extra("tracker"):
    from .mlflow_plugins import LuxonisRequestHeaderProvider
    from .tracker import LuxonisTracker

__all__ = ["LuxonisRequestHeaderProvider", "LuxonisTracker"]



import re

from pydantic import Field, field_validator

from luxonis_ml.utils import BaseModelExtraForbid

from .model import Model

CONFIG_VERSION = "1.0"


class Config(BaseModelExtraForbid):
    """The main class of the multi/single-stage model config scheme
    (multi- stage models consists of interconnected single-stage
    models).

    @type config_version: str
    @ivar config_version: String representing config schema version in
        format 'x.y' where x is major version and y is minor version
    @type model: Model
    @ivar model: A Model object representing the neural network used in
        the archive.
    """

    config_version: str = Field(
        CONFIG_VERSION,
        description="String representing config schema version in format 'x.y' where x is major version and y is minor version.",
    )
    model: Model = Field(
        description="A Model object representing the neural network used in the archive."
    )

    @field_validator("config_version")
    @classmethod
    def validate_config_version_format(cls, v: str) -> str:
        # Regular expression to match 'x.y' where x and y are integers
        if not re.match(r"^\d+\.\d+$", v):
            raise ValueError(
                "'config_version' must be in format 'x.y' where x and y are integers"
            )
        return v



from .base_models import (
    Head,
    HeadMetadata,
    HeadType,
    Input,
    Metadata,
    Output,
    PreprocessingBlock,
)
from .enums import DataType, InputType

__all__ = [
    "DataType",
    "Head",
    "HeadMetadata",
    "HeadType",
    "Input",
    "InputType",
    "Metadata",
    "Output",
    "PreprocessingBlock",
]



import pytest

from luxonis_ml.nn_archive.utils import infer_layout


def test_infer_layout():
    assert infer_layout([1, 3, 256, 256]) == "NCHW"
    assert infer_layout([1, 1, 256, 256]) == "NCHW"
    assert infer_layout([1, 4, 256, 256]) == "NCHW"
    assert infer_layout([1, 19, 7, 8]) == "NCDE"
    assert infer_layout([256, 256, 3]) == "HWC"
    assert infer_layout([256, 256, 1]) == "HWC"
    assert infer_layout([256, 256, 12]) == "HWC"

    with pytest.raises(ValueError, match="Too many dimensions"):
        infer_layout(list(range(30)))



from mlflow import __version__
from mlflow.tracking.request_header.abstract_request_header_provider import (
    RequestHeaderProvider,
)

from luxonis_ml.utils import environ

_USER_AGENT = "User-Agent"
_DEFAULT_HEADERS = {_USER_AGENT: f"mlflow-python-client/{__version__}"}


class LuxonisRequestHeaderProvider(RequestHeaderProvider):
    def in_context(self) -> bool:
        return True

    def request_headers(self) -> dict:
        headers = dict(**_DEFAULT_HEADERS)
        headers["CF-Access-Client-Id"] = environ.MLFLOW_CLOUDFLARE_ID
        headers["CF-Access-Client-Secret"] = environ.MLFLOW_CLOUDFLARE_SECRET
        return headers



from .head import Head, HeadMetadata, HeadType
from .input import Input, PreprocessingBlock
from .metadata import Metadata
from .output import Output

__all__ = [
    "Head",
    "HeadMetadata",
    "HeadType",
    "Input",
    "Metadata",
    "Output",
    "PreprocessingBlock",
]



from functools import lru_cache
from typing import Iterator, Tuple

import numpy as np

from luxonis_ml.typing import Labels, TaskType


@lru_cache
def task_is_metadata(task: str) -> bool:
    """Returns whether a task is a metadata task.

    @type task: str
    @param task: The task to check.
    @rtype: bool
    @return: Whether the task is a metadata task.
    """
    return get_task_type(task).startswith("metadata/")


@lru_cache
def split_task(task: str) -> Tuple[str, str]:
    """Splits a task into its task name and type.

    @type task: str
    @param task: The task to split.
    @rtype: Tuple[str, str]
    @return: A tuple containing the task name and type.
    """
    splits = task.split("/", 1)
    if len(splits) == 1:
        return "", splits[0]
    return splits[0], splits[1]


@lru_cache
def get_task_name(task: str) -> str:
    """Returns the task name from a task.

    @type task: str
    @param task: The task.
    @rtype: str
    @return: The task name.
    """
    return task.split("/")[0]


@lru_cache
def get_task_type(task: str) -> str:
    """Returns the task type from a task.

    Example:

        >>> get_task_type("task_name/type")
        'type'
        >>> get_task_type("metadata/name")
        'metadata/name'
        >>> get_task_type("task_name/metadata/name")
        'metadata/name'

    @type task: str
    @param task: The task in a format like "task_name/type".
    @rtype: str
    @return: The task type. If the task is a metadata task,
        the type will be "metadata/type".
    """
    parts = task.split("/")
    if len(parts) == 1:
        return parts[0]
    if len(parts) == 2:
        if parts[0] == "metadata":
            return task
        return parts[1]
    if parts[-2] == "metadata":
        return f"metadata/{parts[-1]}"
    return task.split("/")[-1]


def task_type_iterator(
    labels: Labels, task_type: TaskType
) -> Iterator[Tuple[str, np.ndarray]]:
    """Iterates over labels of a specific type.

    @type labels: Labels
    @param labels: The labels to iterate over.
    @type task_type: str
    @param task_type: The type of label to iterate over.
    @rtype: Iterator[Tuple[str, np.ndarray]]
    @return: An iterator over the labels of the specified type.
    """
    for task, array in labels.items():
        if get_task_type(task) == task_type:
            yield task, array
